#ifdef CP_CHANGES
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef DEBUG
#include <assert.h>
#endif


#include <stdlib.h>
#include <ctype.h>

#include "rules.h"
#include "decode.h"
#include "debug.h"
#include "plugbase.h"
#include "parser.h"
#include "plugin_enum.h"
#include "util.h"
#include "mstring.h"
#include "event.h"

#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <Policyd.H>
#endif /* !WIN32 */

#ifdef UTM
#include "inline.h"  //For direction marcos

//Parameter to query policyd
typedef struct _QueryPolicyParam
{
    char *ip;
    char *vpnflag;
    POLICYD_ACTION action;
    struct PolicyUserInfo userinfo;
} QueryPolicyParam;
#endif /* UTM */

#ifdef STM
#define POLICYD_TIMEOUT    1
#endif /* STM */

typedef struct _QueryPolicyData
{
    char policy_protocol[20];
    int default_mode;

} QueryPolicyData;

void *instance = NULL;

char *isvpn = "isvpn";

void QueryPolicyInit(char *, OptTreeNode *, int);
void ParseQuery(char *, OptTreeNode *);
int QueryPolicy(Packet *, struct _OptTreeNode *, OptFpList *);

void SetupQueryPolicy(void)
{
    RegisterPlugin("query", QueryPolicyInit);
    
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Plugin: QueryPolicy Initialized\n"););
}

void QueryPolicyInit(char *data, OptTreeNode *otn, int protocol)
{
    Event *event;
    if(otn->ds_list[PLUGIN_QUERYPOLICY])
    {
        FatalError("%s(%d): Multiple query options in rule\n", file_name,
                file_line);
    }

    otn->ds_list[PLUGIN_QUERYPOLICY] = (QueryPolicyData *)
          SnortAlloc(sizeof(QueryPolicyData));

    ParseQuery(data, otn);
    if(instance == NULL)
    {
        POLICYD_RET_CODE ret = createPolicyClientInstance(&instance);
        if(ret != POLICYD_RET_OK)
        {
            instance = NULL;
            ErrorMessage("%s(%d): Create policyd client instance failed, %d\n", 
                          __FILE__, __LINE__, ret);
        }
#ifdef STM
        else
        {
            /* Set timeout value for query operation */
            setPolicydClientTimeout(instance, POLICYD_TIMEOUT);
        }
#endif /* STM */
    }

    event = &(otn->event_data);
    event->userinfo = (struct PolicyUserInfo *)SnortAlloc(sizeof(struct PolicyUserInfo));
    if(event->userinfo == NULL)
    {
        FatalError("%s(%d): Can't Alloc memory for policy user info\n", file_name,
                   file_line);
    }
    
}

void ParseQuery(char *data, OptTreeNode *otn)
{
    QueryPolicyData *ds_ptr;
    char *token, *str, *p;

    ds_ptr = (QueryPolicyData *)otn->ds_list[PLUGIN_QUERYPOLICY];

    while(isspace((int)*data)) data++;
    str = strdup(data);

    if(str == NULL)
    {
        FatalError("ParseQuery: Can't strdup data\n");
    }


    p=str;
    token = strtok(p, ",");

    if(strlen(token) >= 20)
    {
        FatalError("ParseQuery: The protocol name is too long(>=20 bytes");
    }

    strcpy(ds_ptr->policy_protocol, token);

    token = strtok(NULL, ",");
    while(isspace((int)*token)) token++;

    ds_ptr->default_mode = atoi(token);
    if(ds_ptr->default_mode != 0 && ds_ptr->default_mode != 1)
    {
        FatalError("ParseQuery: The default mode is not 0/1"); 
    }

    AddOptFuncToList(QueryPolicy, otn);
    
    free(str);
}

int QueryPolicy(Packet *p, struct _OptTreeNode *otn, OptFpList *fp_list)
{
    QueryPolicyData *ds_ptr;
    POLICYD_ACTION action;
    POLICYD_RET_CODE ret = POLICYD_RET_OK;
    Event *event;

    ds_ptr = (QueryPolicyData *)otn->ds_list[PLUGIN_QUERYPOLICY];

    if(instance == NULL)
    {
        ret = createPolicyClientInstance(&instance);
        if(ret != POLICYD_RET_OK)
        {
            action = POLICYD_ACTION_NO_MATCH;
            instance = NULL;
            ErrorMessage("%s(%d): Create policyd client instance failed, %d\n", 
                          __FILE__, __LINE__, ret);
        }
        else
        {
            action = POLICYD_ACTION_ALLOW;
#ifdef STM
            /* Set timeout value for query operation */
            setPolicydClientTimeout(instance, POLICYD_TIMEOUT);
#endif /* STM */
        }
    }
    else
    {
        action = POLICYD_ACTION_ALLOW;
    }

     
    if(action == POLICYD_ACTION_ALLOW)
    {
        event = &(otn -> event_data);
        if(event->userinfo != NULL)
        {
            /* Init the userinfo and clear the old data  */
            memset(event->userinfo, '\0', sizeof(struct PolicyUserInfo));
        }
#ifdef UTM
        if(p->direction & BETWEEN_SSL_AND_LAN)
        {
            QueryPolicyParam query_param1;
            QueryPolicyParam query_param2;
            memset(&query_param1, '\0', sizeof(QueryPolicyParam));
            memset(&query_param2, '\0', sizeof(QueryPolicyParam));
            query_param1.action = POLICYD_ACTION_UNAUTHORIZED;
            query_param2.action = POLICYD_ACTION_UNAUTHORIZED;

            if(p->direction & TO_WAN)
            {
                /*
                ** Packets from LAN to SSL VPN Tunnel
                ** The source IP is LAN IP.
                ** The dest IP is VPN client IP.
                */
                query_param1.ip = strdup(inet_ntoa(p->iph->ip_src));
                query_param1.vpnflag = NULL;

                query_param2.ip = strdup(inet_ntoa(p->iph->ip_dst));
                query_param2.vpnflag = isvpn;
            }
            else if(p->direction & FROM_WAN)
            {
                /*
                ** Packets from SSL VPN Tunnel to LAN
                ** The source IP is VPN client IP.
                ** The dest IP is LAN IP.
                */
                query_param1.ip = strdup(inet_ntoa(p->iph->ip_src));
                query_param1.vpnflag = isvpn;

                query_param2.ip = strdup(inet_ntoa(p->iph->ip_dst));
                query_param2.vpnflag = NULL;
            }

            ret = queryAppPolicy(instance, ds_ptr->policy_protocol,
                                 query_param1.ip, query_param1.vpnflag,
                                 &(query_param1.action),
                                 &(query_param1.userinfo));
            if(ret != POLICYD_RET_OK)
            {
                query_param1.action = POLICYD_ACTION_UNAUTHORIZED;
                memset(&(query_param1.userinfo), '\0',
                       sizeof(struct PolicyUserInfo));
                ErrorMessage("%s(%d): Query policyd failed, %d.\n",
                             __FILE__, __LINE__, ret);
            }

            if(query_param1.action == POLICYD_ACTION_BLOCK)
            {
                /*
                ** If the action of first query is Block, we don't use the
                ** param2 to query again. And use the action1 as final result.
                */
                query_param2.action = POLICYD_ACTION_UNAUTHORIZED;
                ret = POLICYD_RET_OK;
            }
            else
            {
                //The action1 is not block, use param2 to query.
                ret = queryAppPolicy(instance, ds_ptr->policy_protocol,
                                     query_param2.ip, query_param2.vpnflag,
                                     &(query_param2.action),
                                     &(query_param2.userinfo));
            }

            if(ret != POLICYD_RET_OK)
            {
                query_param2.action = POLICYD_ACTION_UNAUTHORIZED;
                memset(&query_param2.userinfo, '\0',
                       sizeof(struct PolicyUserInfo));
                ErrorMessage("%s(%d): Query policyd failed, %d.\n",
                             __FILE__, __LINE__, ret);
            }

            /*
            ** Select the higher priority action.
            ** high:0 > allow:1 > no_match:2 > unauth:3
            ** Therefor, the action value smaller, the priority is higher.
            */
            if(query_param1.action < query_param2.action)
            {
                action = query_param1.action;
                memcpy(event->userinfo, &(query_param1.userinfo),
                       sizeof(struct PolicyUserInfo));
            }
            else
            {
                action = query_param2.action;
                memcpy(event->userinfo, &(query_param2.userinfo),
                       sizeof(struct PolicyUserInfo));
            }

            if(query_param1.ip)
                free(query_param1.ip);
            if(query_param2.ip)
                free(query_param2.ip);
        }
        else
        {
            QueryPolicyParam query_param;
            memset(&query_param, '\0', sizeof(QueryPolicyParam));

            ////Have the SSL_VPN_PACKET flag, set isvpn flag.
            if(p->direction & SSL_VPN_PACKET)
                query_param.vpnflag = isvpn;

            if(p->direction & TO_WAN)
            {
                /*
                ** Packet from LAN to WAN, or from SSL VPN tunnel to WAN,
                ** or from LAN to SSL VPN port forwarding.
                */
                query_param.ip = strdup(inet_ntoa(p->iph->ip_src));
            }
            else if(p->direction & FROM_WAN)
            {
                /*
                ** Packets from WAN to LAN, or from WAN to SSL VPN tunnel,
                ** or from SSL VPN port forwarding to LAN.
                */
                query_param.ip = strdup(inet_ntoa(p->iph->ip_dst));
            }
            else
            {
                /*
                ** Can't get the direction, maybe the traffic between IPSEC VPN
                ** and WAN.
                */
                query_param.ip = NULL;
                action = POLICYD_ACTION_UNAUTHORIZED;
                memset(event->userinfo, '\0', sizeof(struct PolicyUserInfo));
                ret = POLICYD_RET_OK;
            }

            if(query_param.ip != NULL)
            {
                ret = queryAppPolicy(instance, ds_ptr->policy_protocol,
                                     query_param.ip, query_param.vpnflag,
                                     &action, event->userinfo);
            }

            if(ret != POLICYD_RET_OK)
            {
                action =  POLICYD_ACTION_UNAUTHORIZED;
                memset(event->userinfo, '\0', sizeof(struct PolicyUserInfo));
                ErrorMessage("%s(%d): Query policyd failed, %d.\n",
                            __FILE__, __LINE__, ret);
            }

            if(query_param.ip)
                free(query_param.ip);
         }

#else
        /* 
        ** In STM, we transfer source IP and dest IP to policyd, 
        ** and let policyd to detemine which IP is LAN side.
        */
	
       char *source_ip, *dest_ip;
       source_ip = strdup(inet_ntoa(p->iph->ip_src));
       dest_ip = strdup(inet_ntoa(p->iph->ip_dst));
       ret = queryAppPolicy(instance, ds_ptr->policy_protocol,
                             source_ip, dest_ip, 
                             &action, event->userinfo);
       free(source_ip);
       free(dest_ip);
       if(ret != POLICYD_RET_OK)
       {
           action = POLICYD_ACTION_NO_MATCH;
           ErrorMessage("%s(%d): Query policyd failed, %d.\n", 
                         __FILE__, __LINE__, ret);
       }
#endif /* UTM */
    }

#ifdef DEBUG
    printf("Query match %s, %s, %d, %d\n", ds_ptr->policy_protocol, inet_ntoa(p->iph->ip_src), ds_ptr->default_mode, action);   
#endif 

    if(action == POLICYD_ACTION_BLOCK 
       || (action == POLICYD_ACTION_UNAUTHORIZED && ds_ptr->default_mode == 1)
       || (action == POLICYD_ACTION_NO_MATCH && ds_ptr->default_mode == 1))
    {
        return fp_list->next->OptTestFunc(p, otn, fp_list->next);
    }
    
    return 0; 

}
#endif
